-- 
-- Author: SpiderDave
-- version: 2012.01.29
--

emu_data={}
if (FCEU) then
	emu_data.name="FCEU"
	emu_data.fceu=true
else
	emu_data.name='unknown'
end

--
-- New memory functions
--
function rom   .readword(a) return rom   .readbyte(a) + 256 * rom   .readbyte(a+1) end
function memory.readword(a) return memory.readbyte(a) + 256 * memory.readbyte(a+1) end

function memory.getregisters()
	return memory.getregister("a"),memory.getregister("x"),memory.getregister("y"),memory.getregister("s"),memory.getregister("p"),memory.getregister("pc")
end

function memory.readbyteppu(a)
	memory.writebyte(0x2001,0x00) -- Turn off rendering
	memory.readbyte(0x2002) -- PPUSTATUS (reset address latch)
	memory.writebyte(0x2006,math.floor(a/0x100)) -- PPUADDR high byte
	memory.writebyte(0x2006,a % 0x100) -- PPUADDR low byte
	if a < 0x3f00 then 
		dummy=memory.readbyte(0x2007) -- PPUDATA (discard contents of internal buffer if not reading palette area)
	end
	ret=memory.readbyte(0x2007) -- PPUDATA
	memory.writebyte(0x2001,0x1e) -- Turn on rendering
	return ret
end


function memory.writebyteppu(a,v)
	memory.writebyte(0x2001,0x00) -- Turn off rendering
	memory.readbyte(0x2002) -- PPUSTATUS (reset address latch)
	memory.writebyte(0x2006,math.floor(a/0x100)) -- PPUADDR high byte
	memory.writebyte(0x2006,a % 0x100) -- PPUADDR low byte
	memory.writebyte(0x2007,v) -- PPUDATA
	memory.writebyte(0x2001,0x1e) -- Turn on rendering
end

-- Write a string of bytes to memory
function memory.writebytes(address,str)
	local i
	local c
	for i = 0, #str-1 do
		memory.writebyte(address+i,string.byte(str,i+1))
	end
end

--fonts={}
--fonts[letters]="ABCDEFGHIJKLMNOPQRSTUVWXYZ.,?!abcdefghijklmnopqrstuvwxyz-*@#0123456789:&''\"             "
--font_letters="ABCDEFGHIJKLMNOPQRSTUVWXYZ.,?!abcdefghijklmnopqrstuvwxyz-*@#0123456789:&''\"             "

joypad_data={[1]=joypad.read(1), [2]=joypad.read(2)}
function joypad_read()
	joypad_data['old']={joypad_data[1],joypad_data[2]}
	-- button_names={[1]='up',[2]='down',[3]='left',[4]='right',[5]='A',[6]='B',[7]='select',[8]='start'}
	button_names={'up','down','left','right','A','B','select','start'}
	
	for i=1,2 do
		joypad_data[i]=joypad.read(i)
		for ii=1,8 do
			b=button_names[ii]
			joypad_data[i][b..'_press']=(joypad_data[i][b] and not joypad_data.old[i][b])
			joypad_data[i][b..'_press_time']=joypad_data[i][b] and math.min((joypad_data.old[i][b..'_press_time'] or 0) +1,1000) or 0
			joypad_data[i]['idle']= (joypad_data[i].up~=true and joypad_data[i].down~=true and joypad_data[i].left~=true and joypad_data[i].right~=true and joypad_data[i].A~=true and joypad_data[i].B~=true and joypad_data[i].select~=true and joypad_data[i].start~=true)
			joypad_data[i]['idle_time']=joypad_data[i]['idle'] and math.min((joypad_data.old[i]['idle_time'] or 0) +1,1000) or 0
			truefalse={[true]='true',[false]='false'}
			--[[
			gui.text(8, 8+8*2+8, string.format("%s",truefalse[joypad_data[1].idle]) ,"white","clear")
			gui.text(8, 8+8*2+8*ii, string.format("%s %s",ii, b) ,"white","clear")
			gui.text(8+50, 8+8*2+8*ii, string.format("%s",joypad_data[i][string.format('%s_press', b)] and 1 or 0) ,"white","clear")
			gui.text(8+100, 8+8*2+8*ii, string.format("%s",joypad_data[i][b..'_press_time']) ,"white","clear")
			]]--
		end
	end
	
	return joypad_data
end

--
-- A drop-in replacement for input.read() with more options.
--
-- leftbutton_press returns true the moment left button is pressed
-- leftbutton_release returns true the moment left button is released
-- leftbutton_click returns true when the left button is pressed and released without movement
-- selection.x selection.y selection.width selection.height
--
--
input_data={['current']=input.read()}
function input_read()
	input_data['old']=input_data.current
	input_data.current=input.read()
	input_data.current.selection={}
	
	--depreciated; use leftbutton_press
	input_data.current.leftclick_real=(input_data.current.leftclick and not input_data.old.leftclick)
	
	input_data.current.leftbutton=input_data.current.leftclick
	input_data.current.middlebutton=input_data.current.middleclick
	input_data.current.rightbutton=input_data.current.rightclick
	
	input_data.current.leftbutton_press=(input_data.current.leftclick and not input_data.old.leftclick)
	input_data.current.leftbutton_release=(not input_data.current.leftclick and input_data.old.leftclick)
	
	input_data.current.middlebutton_press=(input_data.current.middleclick and not input_data.old.middleclick)
	input_data.current.middlebutton_release=(not input_data.current.middleclick and input_data.old.middleclick)
	
	input_data.current.rightbutton_press=(input_data.current.rightclick and not input_data.old.rightclick)
	input_data.current.rightbutton_release=(not input_data.current.rightclick and input_data.old.rightclick)
	
	if input_data.current.leftbutton_press then
		input_data.current.xmouse_down=input_data.current.xmouse
		input_data.current.ymouse_down=input_data.current.ymouse
	else
		input_data.current.xmouse_down=input_data.old.xmouse_down
		input_data.current.ymouse_down=input_data.old.ymouse_down
	end
	
	-- snap 8
	if input_data.current.xmouse_down and input_data.current.ymouse_down then
		input_data.current.xmouse_down=input_data.current.xmouse_down-input_data.current.xmouse_down % 8
		input_data.current.ymouse_down=input_data.current.ymouse_down-input_data.current.ymouse_down % 8
	end
	
	if input_data.current.xmouse==input_data.old.xmouse and input_data.current.ymouse==input_data.old.ymouse then
		input_data.current.mouseidle=math.min((input_data.old.mouseidle or 0) +1,1000)
	else
		input_data.current.mouseidle=0
	end
	
	if input_data.current.leftbutton_release then
		if input_data.current.xmouse==input_data.current.xmouse_down and input_data.current.ymouse==input_data.current.ymouse_down then
			-- Click is pressing and releasing without movement
			input_data.current.leftbutton_click=true
		end
		input_data.current.xmouse_up=input_data.current.xmouse
		input_data.current.ymouse_up=input_data.current.ymouse
		
		-- snap 8
		if input_data.current.xmouse_up and input_data.current.ymouse_up then
			input_data.current.xmouse_up=(input_data.current.xmouse_up+4)-(input_data.current.xmouse_up+4) % 8
			input_data.current.ymouse_up=(input_data.current.ymouse_up+4)-(input_data.current.ymouse_up+4) % 8
		end
		
		if input_data.current.xmouse_up>input_data.current.xmouse_down then
			input_data.current.selection.x=input_data.current.xmouse_down
			input_data.current.selection.width=input_data.current.xmouse_up-input_data.current.xmouse_down
		else
			input_data.current.selection.x=input_data.current.xmouse_up
			input_data.current.selection.width=input_data.current.xmouse_down-input_data.current.xmouse_up
		end
		if input_data.current.ymouse_up>input_data.current.ymouse_down then
			input_data.current.selection.y=input_data.current.ymouse_down
			input_data.current.selection.height=input_data.current.ymouse_up-input_data.current.ymouse_down
		else
			input_data.current.selection.y=input_data.current.ymouse_up
			input_data.current.selection.height=input_data.current.ymouse_down-input_data.current.ymouse_up
		end
	else
		input_data.current.xmouse_up=input_data.old.xmouse_up
		input_data.current.ymouse_up=input_data.old.ymouse_up
	end

	if input_data.showselect and input_data.current.leftclick then
		x2=input_data.current.xmouse
		y2=input_data.current.ymouse
		-- snap 8
		x2=(x2+4)-(x2+4) % 8
		y2=(y2+4)-(y2+4) % 8
		gui.text(input_data.current.xmouse_down,input_data.current.ymouse_down-8,string.format('(%s,%s) %sx%s',input_data.current.xmouse_down, input_data.current.ymouse_down,math.abs(x2-input_data.current.xmouse_down),math.abs(y2-input_data.current.ymouse_down)), "white", "black")
		gui.drawbox(input_data.current.xmouse_down, input_data.current.ymouse_down, x2, y2, 'clear',"white")
	end

	return input_data.current
end

function input_showselect(show)
	if show==true then
		input_data.showselect=true
	else
		input_data.showselect=nil
	end
end

-- load a font from .gd files.  DEPRECIATED
function loadfont(path)
	-- takes path to font files, ending in \
	font={}
	str="ABCDEFGHIJKLMNOPQRSTUVWXYZ"
	--str=font.letters
	for i = 1, #str do
		local c = str:sub(i,i)
		local file = io.open(path..c..'.gd',"rb")
		io.input(file)
		font[c]=io.read('*a')
		io.close(file)
	end
	return font
end

function getfilecontents(path)
	local file = io.open(path,"rb")
	if file==nil then return nil end
	io.input(file)
	ret=io.read('*a')
	io.close(file)
	return ret
end

function writetofile(path,stuff)
	if not stuff then return nil end
	local file = io.open(path,"wb")
	io.output(file)
	io.write(stuff)
	io.close(file)
	return true
end


-- draw font loaded with loadfont function
function drawfont(x,y,font,str)
	local originx=x
	local originy=y
	if (not font) then return false end
	for i = 1, #str do
		local c = str:sub(i,i)
		if (c=='\n') then
			x=originx
			y=y+8
		elseif (font[c]) then
			gui.gdoverlay(x,y, font[c])
			x=x+8
		end
	end
end

function bin2hex(str)
	local output=''
	for i = 1, #str do
		local c = string.byte(str:sub(i,i))
		output=output..string.format("%02x", c)
	end
	return output
end

function hex2bin(str)
	local output=''
	for i = 1, (#str/2) do
		local c = str:sub(i*2-1,i*2)
		output=output..string.char(tonumber(c, 16))
	end
	return output
end

--[[--------------------------------------------------
function to display memory on screen; good for debugging n stuff
    example usage:
        showmem(0x1234)  --use this ONLY if you're using address alone
        showmem{address=0x1234, x=0,y=0} --note the special {} syntax!
        showmem{0x1234, 0, 0} --this works too
    parameters:
        address, x, y
--------------------------------------------------]]--
function showmem(args)
	if not args then args={} end
	if (type(args)=="number") then
		address=args
		args={}
	else
		address=args.address or args[1] or 0
	end
	
	local _address,x,y
	
	x=args.x or args[2] or 8
	y=args.y or args[3] or 8
	
	_address=address
	
	for j=0,5-1 do
		for i=0,16-1 do
			--_address=address+j*(16*0x20)+0x20*i
			_address=_address+1
			gui.text(x+8*(6*j), y+8*(1+i), string.format('%X %02X  ',_address, memory.readbyte(_address)) ,"white","black")
		end
	end
end

-- Save font as lua code
function savefont_x(filename, font_table)
	--font={}
	local fontnum
	out='font={}\n'
	out=out.."font['letters']=\"ABCDEFGHIJKLMNOPQRSTUVWXYZ.,?!abcdefghijklmnopqrstuvwxyz-*@#0123456789:&''\\".. '"$;           "\n'
	for fontnum= 1,99 do
		if not (font_table[fontnum]) then break end
		out=out..string.format("font[%s]={\n",fontnum)
		if not (font_table[fontnum]['"']) then font_table[fontnum]['"']=font_table[fontnum][''] end
		for i = 1, #font_table.letters do
			local c = font_table.letters:sub(i,i)
			--io.input(file)
			if (font_table[fontnum][c]) then
				if (c=="'") then
					out=out..string.format("['\\%s']=hex2bin('%s'),\n",c, bin2hex(font_table[fontnum][c])  )
				else
					out=out..string.format("['%s']=hex2bin('%s'),\n",c, bin2hex(font_table[fontnum][c])  )
				end
			end
			--font[c]=io.read('*a')
		end
		out=out:sub(1,#out-2).."\n" --remove last comma
		out=out..'}\n'
	end
	print(out)

	local file = io.open(filename,"w")
	io.output(file)
	io.write(out)
	io.close(file)
end


function showcolor(x,y)
	--emu.getscreenpixel(x,y,true)
	gui.text(8*25, 8*1, string.format('%02X %02X %02X ',emu.getscreenpixel(x,y,true)) ,"white","black")
end

-- Create a gd string from a section of the screen
_palette={}
function gdfromscreen(x,y,w,h,opt)
	if w<1 or h<1 then return nil end
	if not opt then opt={} end
	if opt.transparent==nil then opt.transparent=true end
	if opt.transparentcolor==nil then opt.transparentcolor='#000000' end
	if opt.nobk==nil then opt.nobk=false end
	
	local gdstr=''
	local _mx=x
	local _my=y
	--if opt.nobk then emu.setrenderplanes(true, false); emu.frameadvance() end
	
	-- We're going to mess with the vram and advance it a little so we'll just undo when done.
	save1=savestate.object()
	savestate.save(save1)
	if opt.nobk then
		for _i=0,0x10-1 do
			memory.writebyteppu(0x3f00+_i,0x15)
		end
		emu.frameadvance()
	end
	
	for y = 0, h-1 do
		for x = 0, w-1 do
			local t=0
			local r, g, b=emu.getscreenpixel(_mx+x,_my+y+0,true)
			--if opt.transparent==true and r+g+b==0 then t=0x80 end
			--if opt.transparent==true and r==0xe4 and g==0x00 and b==0x58 then r=0;g=0;b=0;t=0x80 end
			if opt.transparent==true and string.format('#%02X%02X%02X',r,g,b)==opt.transparentcolor then r=0;g=0;b=0;t=0x80 end
			gdstr=gdstr..string.format('%02x%02x%02x%02x',t,r,g,b)
		end
	end
	--if opt.nobk then emu.setrenderplanes(true, true) end
	savestate.load(save1)
	save1=nil
	gdstr=hex2bin('FFFE00'..string.format("%02x", w)..'00'..string.format("%02x", h)..'01'..'FFFFFFFF'..gdstr)
	return gdstr
end


yesno={[true]="yes",[false]="no"}

waitforrelease=false
--Create a button with text and an optional toggle value (shows "on" or "off")
--Call it in a loop, returns true when clicked.
function button(x,y,txt,toggle)
	if input_data.current.mouseidle>80 then return false end
	ret=false
	xw=96
	yw=8
	c="white"
	c2="black"
	if(input_data.current.xmouse>x and input_data.current.xmouse<x+xw and input_data.current.ymouse>y and input_data.current.ymouse<y+yw) then
		c="white"
		c2="blue"
		--if (waitforrelease) then
		--	c="white"
		--	c2="cyan"
		--end
		if input_data.current.leftbutton_press then
		--if (inp.leftclick and waitforrelease==false) then
			c="blue"
			c2="white"
			ret=true
			--waitforrelease=true
		--else
			--if not (inp.leftclick) then waitforrelease=false end
		end
	end
	gui.drawbox(x, y, x+xw, y+yw, c2, c2)
	onoff={[true]="on",[false]="off"}
	--txt=string.format("%s %s", txt, onoff[toggle])
	gui.text(x, y, txt,c,"clear")
	if not (toggle==nil) then
		gui.text(x+xw-16, y, onoff[toggle] ,c,"clear")
	end
	return ret
end


--------------------------------------------------------------------------------
--
-- Create gd image out of an 8x8, 4-color tile in a pattern table
--
function gdTile(ofs,c0,c1,c2,c3,hflip,double)
    local gd = "\255\254\0\008\0\008\001\255\255\255\255"
    if double then gd = "\255\254\0\016\0\016\001\255\255\255\255" end
    --memory.writebyte(0x2001,0x00) -- Turn off rendering
    local v0,v1
    for y=0,7 do
        --local v0 = rom.readbyte(ofs + y    )
        --local v1 = rom.readbyte(ofs + y + 8)
        --local v0 = memory.readbyteppu(ofs + y    )
        --local v1 = memory.readbyteppu(ofs + y + 8)
        memory.writebyte(0x2006,math.floor((ofs+y)/0x100)) -- PPUADDR high byte
        memory.writebyte(0x2006,(ofs + y) % 0x100) -- PPUADDR low byte
        dummy = memory.readbyte(0x2007) -- PPUDATA (discard internal buffer contents)
        v0 = memory.readbyte(0x2007) -- PPUDATA
        memory.writebyte(0x2006,math.floor((ofs+y+8)/0x100)) -- PPUADDR high byte
        memory.writebyte(0x2006,(ofs + y+8) % 0x100) -- PPUADDR low byte
        dummy = memory.readbyte(0x2007) -- PPUDATA (discard internal buffer contents)
        v1 = memory.readbyte(0x2007) -- PPUDATA
        
        local line = ""
        if hflip then
            for x=0,7 do
                local px
                if AND(v1,1) ~= 0 then
                    if AND(v0,1) ~= 0 then
                        v0 = v0 - 128
                        px = c3
                    else
                        px = c2
                    end
                else
                    if AND(v0,1) ~= 0 then
                        px = c1
                    else
                        px = c0
                    end
                end
                line = line .. px
                if double then line = line .. px end
                v1 = math.floor(v1/2)
                v0 = math.floor(v0/2)
            end
        else
            for x=0,7 do
                if v1 >= 128 then
                    v1 = v1 - 128
                    if v0 >= 128 then
                        v0 = v0 - 128
                        px = c3
                    else
                        px = c2
                    end
                else
                    if v0 >= 128 then
                        v0 = v0 - 128
                        px = c1
                    else
                        px = c0
                    end
                end
                line = line .. px
                if double then line = line .. px end
                v1 = v1 * 2
                v0 = v0 * 2
            end
        end
        gd = gd .. line
        if double then gd = gd .. line end
    end
    --memory.writebyte(0x2001,0x1e) -- Turn on rendering
    return gd
end


function getpalettedata()
	local _palettedata={}
	_palettedata.color_indexed={}
	local _i
	for _i=0,25-1 do
		_palettedata.color_indexed[_i]=memory.readbyteppu(0x3f00+_i)
	end
	return _palettedata
end

nespalette={[0]=
'#7C7C7C',
'#0000FC',
'#0000BC',
'#4428BC',
'#940084',
'#A80020',
'#A81000',
'#881400',
'#503000',
'#007800',
'#006800',
'#005800',
'#004058',
'#000000',
'#000000',
'#000000',
'#BCBCBC',
'#0078F8',
'#0058F8',
'#6844FC',
'#D800CC',
'#E40058',
'#F83800',
'#E45C10',
'#AC7C00',
'#00B800',
'#00A800',
'#00A844',
'#008888',
'#000000',
'#000000',
'#000000',
'#F8F8F8',
'#3CBCFC',
'#6888FC',
'#9878F8',
'#F878F8',
'#F85898',
'#F87858',
'#FCA044',
'#F8B800',
'#B8F818',
'#58D854',
'#58F898',
'#00E8D8',
'#787878',
'#000000',
'#000000',
'#FCFCFC',
'#A4E4FC',
'#B8B8F8',
'#D8B8F8',
'#F8B8F8',
'#F8A4C0',
'#F0D0B0',
'#FCE0A8',
'#F8D878',
'#D8F878',
'#B8F8B8',
'#B8F8D8',
'#00FCFC',
'#F8D8F8',
'#000000',
'#000000'
}

menu={
x=72,
y=88,
background=true,
index=0,
items={},
background_color='black',

doinput=function()
	if (joypad_data[1]['up_press'] or (joypad_data[1]['up_press_time'] > 20 and joypad_data[1]['up_press_time'] % 5 ==0)) then
		menu.index=menu.index-1
		if (menu.index<0) then menu.index=0 end
	end
	if (joypad_data[1]['down_press'] or (joypad_data[1]['down_press_time'] > 20 and joypad_data[1]['down_press_time'] % 5 ==0)) then
		menu.index=menu.index+1
		if (menu.index>#menu.items) then menu.index=#menu.items end
	end
	if joypad_data[1]['A_press'] then
		menu.items[menu.index].action()
	end
end,

show=function()
	if menu.background==true then gui.drawbox(0, 0, 256, 240, menu.background_color, menu.background_color) end
	menu.text=''
	local _i=0
	for _i=0,#menu.items do
		drawfont(menu.x,menu.y+8*_i,menu.font, menu.items[_i].text)
	end
	
	-- Display custom cursor image if set, otherwise use a blinking '-'
	if menu.cursor_image then
		gui.gdoverlay(menu.x-12,menu.y+8*menu.index,menu.cursor_image)
	else
		if (emu.framecount() % 24>12) then
			drawfont(menu.x-12,menu.y+8*menu.index,menu.font, "-") --cursor
		end
	end
	
end

}